# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuffers

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Text(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsText(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Text()
        x.Init(buf, n + offset)
        return x

    # Text
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Text
    def Position(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from casymda.visualization.web_server.flatbuffers.CanvasPosition import CanvasPosition
            obj = CanvasPosition()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Text
    def Anchor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from casymda.visualization.web_server.flatbuffers.Anchor import Anchor
            obj = Anchor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Text
    def Text(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Text
    def Fill(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Text
    def FontFamily(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Text
    def FontSize(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 10

def TextStart(builder): builder.StartObject(6)
def TextAddPosition(builder, position): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(position), 0)
def TextAddAnchor(builder, anchor): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(anchor), 0)
def TextAddText(builder, text): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(text), 0)
def TextAddFill(builder, fill): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(fill), 0)
def TextAddFontFamily(builder, fontFamily): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(fontFamily), 0)
def TextAddFontSize(builder, fontSize): builder.PrependUint8Slot(5, fontSize, 10)
def TextEnd(builder): return builder.EndObject()



import math
import operator
from typing import Any, List, Optional

import simpy
from casymda.blocks.block_components import VisualizableBlock
from .coordinates_holder import CoordinatesHolder
from casymda.blocks.tilemap.segment import Segment
from casymda.environments.realtime_environment import (
    ChangeableFactorRealtimeEnvironment,
)
from casymda.visualization.tilemap.tilemap_visualizer_interface import (
    TilemapVisualizerInterface as TilemapVisualizer,
)
from simpy import RealtimeEnvironment


class TilemapMovement(VisualizableBlock):
    """
    Simulates and animates tilemap movements from a start node to a target node,
    either in one timestep (no visualizer set),
    or in steps of a configurable length (visualizer set).
    """

    MAX_ANIMATION_TIMESTEP = 0.1

    def __init__(
        self,
        env,
        name,
        xy=None,
        speed=1,  # px / sec (sim-time)
        coordinates_holder=None,
        from_node="",
        to_node="",
        ways={},
        destroy_anim_on_arrival=True,
        block_capacity=float("inf"),
    ):
        super().__init__(env, name, xy=xy, ways=ways, block_capacity=block_capacity)

        self.speed = speed
        if coordinates_holder is None:
            raise Exception("coordinates holder not set in block: " + name)
        self.coordinates_holder: CoordinatesHolder = coordinates_holder
        self.from_node = from_node
        self.to_node = to_node

        self.tilemap_visualizer: Optional[TilemapVisualizer] = None
        self.destroy_anim_on_arrival = destroy_anim_on_arrival

    def actual_processing(self, entity):
        (
            path_coords,
            overall_length,
        ) = self.coordinates_holder.get_path_coords_and_length_from_to(
            self.from_node, self.to_node
        )

        speed = self.speed

        animation_loop: Any = None
        if self.tilemap_visualizer is not None:
            animation_loop = self.env.process(
                self.animation_loop(
                    entity,
                    path_coords,
                    destroy_on_arrival=self.destroy_anim_on_arrival,
                    speed=speed,
                )  # run animation seperately if required
            )

        time_needed = overall_length / speed
        yield self.env.timeout(time_needed)

        if self.tilemap_visualizer is not None:
            animation_loop.interrupt()  # stop animation running "in parallel"

    def animation_loop(
        self, entity, path_coords, destroy_on_arrival: bool = True, speed=None
    ):
        if speed is None:
            speed = self.speed

        self.tilemap_visualizer.destroy(entity)  # recreate possibly existing animations
        segments = self.get_segments(path_coords, speed)
        time_spent = 0
        try:
            while True:  # while not interupted:
                self.calc_progress_and_animate(entity, segments, time_spent, speed)

                self.set_animation_timestep()
                yield self.env.timeout(self.MAX_ANIMATION_TIMESTEP)
                time_spent += self.MAX_ANIMATION_TIMESTEP
        except simpy.Interrupt:
            """looks like we arrived"""
            x, y = path_coords[-1]
            self.tilemap_visualizer.animate(entity, x, y, self.env.now)
            if destroy_on_arrival:
                self.tilemap_visualizer.destroy(entity)

    def calc_progress_and_animate(
        self, entity, segments: List[Segment], time_spent, speed
    ):
        distance = speed * time_spent
        current_segment = segments[-1]
        for segment in segments:
            current_segment = segment
            if segment.cumulated_length > distance:
                break
        remaining_length = max(current_segment.cumulated_length - distance, 0)
        current_length = current_segment.length - remaining_length
        progress = (
            (
                tuple(
                    map(
                        lambda x: x * (current_length / current_segment.length),
                        current_segment.direction,
                    )
                )
            )
            if current_segment.length > 0
            else current_segment.direction
        )
        current_position = tuple(map(operator.add, progress, current_segment.origin))
        x, y = current_position
        self.tilemap_visualizer.animate(entity, x, y, self.env.now)

    def set_animation_timestep(self):
        if isinstance(self.env, ChangeableFactorRealtimeEnvironment) or isinstance(
            self.env, RealtimeEnvironment
        ):
            self.MAX_ANIMATION_TIMESTEP = 0.04 / self.env.factor
            # keep reference to the factor -> change animation timestep dynamically if visualized (for a smooth animation)

    def get_segments(self, path_coords, speed):
        """pre-compute needed segment information for the animation"""
        segments = []
        cumulated_length = 0
        cumulated_time = 0
        for pc_idx in range(len(path_coords) - 1):
            origin = path_coords[pc_idx]
            destination = path_coords[pc_idx + 1]
            direction = tuple(map(operator.sub, destination, origin))
            squared = map((lambda x: x ** 2), direction)
            length = math.sqrt(sum(squared))
            time_needed = length / speed

            cumulated_length += length
            cumulated_time += time_needed

            segment = Segment(
                origin, destination, direction, length, cumulated_length, cumulated_time
            )
            segments.append(segment)
        return segments



"""seize resource"""
from casymda.blocks.block_components import VisualizableBlock
from casymda.blocks import Entity


class ResourceSeizeQueue(VisualizableBlock):
    """waiting for free resource before seizing and moving forward"""

    def __init__(self, env, name, xy=None, ways=None, resource=None):
        super().__init__(env, name, xy=xy, ways=ways)

        self.resource_to_request = resource

    def actual_processing(self, entity: Entity):
        req = self.resource_to_request.request()
        yield req
        entity.seized_resources.append(req)



"""runs simulation model in its own process"""

import json
from ctypes import c_bool
from multiprocessing import Manager, Process, Value
from typing import Optional

from casymda.environments.realtime_environment import SyncedFloat
from casymda.visualization.canvas import web_canvas

from .flatbuffers import flatbuffer_serializer


class RunnableSimulation:
    """abstract base class for simulations to be run via a SimController"""

    width: int
    height: int
    root_file: str

    def simulate(
        self, shared_state: dict, should_run: Value, factor: SyncedFloat
    ) -> None:
        raise NotImplementedError("abc")


class SimController:
    """wraps RunnableSimulation to be controlled by a sim-server"""

    def __init__(self, simulation: RunnableSimulation) -> None:

        self.simulation = simulation

        self.sim_process: Optional[Process] = None
        self.shared_state: dict = Manager().dict()
        self.should_run = Value(c_bool, False)
        self.factor: SyncedFloat = SyncedFloat._create_factor_instance()

    def start_simulation_process(self):
        self.reset_sim()
        self._setup_sim()
        self.sim_process.start()
        return "simulation process started"

    def _setup_sim(self):
        self.shared_state = Manager().dict()
        self.should_run.value = True
        self.sim_process = Process(
            target=self.simulation.simulate,
            args=(
                self.shared_state,
                self.should_run,
                self.factor,
            ),
        )

    def pause_sim(self):
        self.should_run.value = False
        return "paused"

    def resume_sim(self):
        self.should_run.value = True
        return "resumed"

    def reset_sim(self):
        if self.sim_process is not None and self.sim_process.is_alive():
            self.sim_process.terminate()
            self.sim_process = None
        self.should_run.value = False
        return "reset"

    def get_state_dumps(self):
        state = self._get_state_dumps()
        return json.dumps(state)

    def get_partial_state_dumps(self):
        state = self._get_partial_state_dumps()
        return json.dumps(state)

    def get_state_dumps_fb(self):
        state = self._get_state_dumps()
        return flatbuffer_serializer.serialize(state)

    def get_partial_state_dumps_fb(self):
        state = self._get_partial_state_dumps()
        return flatbuffer_serializer.serialize(state)

    def _get_state_dumps(self):
        self.shared_state[web_canvas.UPDATED_KEY] = set()  # reset
        state = self.shared_state.copy()
        del state[web_canvas.UPDATED_KEY]
        return state

    def _get_partial_state_dumps(self):
        updated_set = set(self.shared_state[web_canvas.UPDATED_KEY])
        state = self.shared_state.copy()
        self.shared_state[web_canvas.UPDATED_KEY] = set()  # reset
        del state[web_canvas.UPDATED_KEY]
        for not_updated in state.keys() - updated_set:
            del state[not_updated]
        return state

    def set_rt_factor(self, value: float):
        return "set factor to " + str(self.factor.set_value(value))

    def get_sim_width(self) -> int:
        return self.simulation.width

    def get_sim_height(self) -> int:
        return self.simulation.height



from casymda.blocks.entity import Entity


class TilemapVisualizerInterface:
    """ interface for tilemap visualizations """

    def animate(self, entity: Entity, x: float, y: float, current_time: float):
        raise NotImplementedError()

    def destroy(self, entity: Entity):
        raise NotImplementedError()



from typing import List, Optional, Callable

from simpy import Environment, Resource

from casymda.blocks.block_components.state import StateManager, States
from casymda.blocks.entity import Entity


class Block:
    """
    Provides simulation-specific behavior of a basic block.
    Entities are accepted via the process_entity entry-point, and standard processing steps are executed.
    Actual processing behavior is to be provided by subclasses, implementing the actual_processing method.
    """

    def __init__(self, env: Environment, name: str, block_capacity=float("inf")):
        self.name = name
        self.env: Environment = env
        self.overall_count_in = 0
        on_enter_or_exit_method_callable = Callable[
            [Entity, Optional[Block], Optional[Block]], None
        ]
        self.do_on_enter_list: List[on_enter_or_exit_method_callable] = []
        self.do_on_exit_list: List[on_enter_or_exit_method_callable] = []
        self.entities: List[Entity] = []
        self.successors: List[Block] = []
        self.block_resource = Resource(env=env, capacity=block_capacity)
        self.state_manager = StateManager(self.on_block_state_change)

        self.env.process(self.late_state_evaluation())

    def on_enter(self, entity: Entity):
        """called when process_entity starts"""
        for method in self.do_on_enter_list:
            method(entity, None, self)

    def on_exit(self: "Block", entity: Entity, successor: "Block"):
        """called when an entities leaves the block"""
        self.entities.remove(entity)
        for method in self.do_on_exit_list:
            method(entity, self, successor)
        self.on_block_change()
        self.on_entity_movement(entity, successor)

    def process_entity(self, entity: Entity):
        """main entry point for entities coming from predecessors"""
        entity.time_of_last_arrival = self.env.now
        self.on_enter(entity)
        self.overall_count_in += 1
        yield self.env.process(self._process_entity(entity))

    def _process_entity(self, entity: Entity):
        """internal entry point for start of entity processing without calling on_enter or setting entity.time_of_last_arrival.
        useful e.g. for wip init"""
        self.entities.append(entity)
        self.on_block_change()

        # processing
        self.state_manager.increment_state_count(States.busy)
        entity.current_process = self.env.process(self.actual_processing(entity))
        yield entity.current_process  # might be used for interrupt

        # wait for successor
        successor = self.find_successor(entity)
        req = successor.block_resource.request()
        self.state_manager.increment_state_count(States.blocked)
        self.state_manager.decrement_state_count(States.busy)
        yield req  # wait until the chosen successor is ready

        # leaving
        self.block_resource.release(entity.block_resource_request)
        entity.block_resource_request = req  # remember to be released

        self.on_exit(entity, successor)
        self.state_manager.decrement_state_count(States.blocked)

        self.env.process(successor.process_entity(entity))

    def find_successor(self, entity: Entity):
        """find next block to send entity to"""
        return self.successors[0]

    def on_block_change(self):
        """called on block change"""

    def on_block_state_change(self, state, new_value):
        """called when state of the block changes"""

    def on_entity_movement(self, entity: Entity, successor: "Block"):
        """called on entity movement from this block to the successor"""

    def late_state_evaluation(self):
        """schedule evaluation on sim start,
        when the visualizer has been loaded"""
        yield self.env.timeout(0)
        self.state_manager.evaluate_state_count()

    def actual_processing(self, entity: Entity):
        """to be implemented by concrete subclasses"""
        raise NotImplementedError()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuffers

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Photo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPhoto(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Photo()
        x.Init(buf, n + offset)
        return x

    # Photo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Photo
    def Factor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 1.0

    # Photo
    def Path(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def PhotoStart(builder): builder.StartObject(2)
def PhotoAddFactor(builder, factor): builder.PrependFloat32Slot(0, factor, 1.0)
def PhotoAddPath(builder, path): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(path), 0)
def PhotoEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuffers

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Image(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsImage(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Image()
        x.Init(buf, n + offset)
        return x

    # Image
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Image
    def Position(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from casymda.visualization.web_server.flatbuffers.CanvasPosition import CanvasPosition
            obj = CanvasPosition()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Image
    def Anchor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from casymda.visualization.web_server.flatbuffers.Anchor import Anchor
            obj = Anchor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Image
    def PhotoId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Image
    def Path(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Image
    def Factor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 1.0

def ImageStart(builder): builder.StartObject(5)
def ImageAddPosition(builder, position): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(position), 0)
def ImageAddAnchor(builder, anchor): builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(anchor), 0)
def ImageAddPhotoId(builder, photoId): builder.PrependUint16Slot(2, photoId, 0)
def ImageAddPath(builder, path): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(path), 0)
def ImageAddFactor(builder, factor): builder.PrependFloat32Slot(4, factor, 1.0)
def ImageEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuffers

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class CanvasPosition(object):
    __slots__ = ['_tab']

    # CanvasPosition
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # CanvasPosition
    def X(self): return self._tab.Get(flatbuffers.number_types.Uint16Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # CanvasPosition
    def Y(self): return self._tab.Get(flatbuffers.number_types.Uint16Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(2))

def CreateCanvasPosition(builder, x, y):
    builder.Prep(2, 4)
    builder.PrependUint16(y)
    builder.PrependUint16(x)
    return builder.Offset()



from typing import Dict, List, Optional, Tuple

from casymda.blocks.block_components.block import Block
from casymda.blocks.block_components.visualizer_interface import BlockVisualizer
from simpy import Environment


class VisualizableBlock(Block):
    """ extends block by visualization-related behavior """

    def __init__(
        self,
        env: Environment,
        name: str,
        block_capacity=float("inf"),
        xy: Tuple[int, int] = (0, 0),
        ways: Dict[str, List[Tuple[int, int]]] = {},
    ):

        super().__init__(env, name, block_capacity=block_capacity)

        self.visualizer: Optional[BlockVisualizer] = None

        self.ways = ways
        self.xy_position: Tuple[int, int] = xy

        self.queuing: bool = True  # read by visualizer

    def on_block_change(self):
        """ animate the block, including e.g. it's contents """
        super().on_block_change()
        if self.visualizer is not None:
            # to be done: the extension of Block could be replaced by
            # a composition and provision of the callbacks
            self.visualizer.animate_block(self)

    def on_block_state_change(self, state, new_value):
        super().on_block_state_change(state, new_value)
        if self.visualizer is not None:
            self.visualizer.change_block_state(self, state, new_value)

    def on_entity_movement(self, entity, successor):
        super().on_entity_movement(entity, successor)
        if self.visualizer is not None:
            self.visualizer.animate_entity_flow(entity, self, successor)



"""canvas based on tkinter"""
import platform
from tkinter import Tk, Canvas

from PIL import Image, ImageTk

from casymda.visualization.canvas.scaled_canvas import ScaledCanvas


class ScaledCanvasTk(ScaledCanvas):
    """draws onto tkinter canvas"""

    def __init__(self, tk_gui: Tk, width, height, scale=1.0):
        super().__init__(scale)
        self.tk_gui = tk_gui
        self.width = width
        self.height = height
        self.canvas = Canvas(tk_gui, width=self.width, height=self.height, bg="white")
        self.canvas.pack(side="left")
        self.update()

    def load_image_file(self, path: str):
        """takes file path and returns loaded image file reference"""

        image = (
            Image.open(path).convert("RGBA")
            if platform.system() != "Darwin"
            else Image.open(path).convert("RGB")
        )
        width, height = self._scale_coords(image.size)

        resized_image = image.resize((width, height), Image.ANTIALIAS)
        image_file = ImageTk.PhotoImage(resized_image)

        return image_file

    def create_image(self, x_coord: int, y_coord: int, image_file, anchor="c"):
        x_y = self._scale_coords((x_coord, y_coord))
        image = self.canvas.create_image(*x_y, image=image_file, anchor=anchor)
        self.update()
        return image

    def create_text(
        self,
        x_coord: int,
        y_coord: int,
        text: str = "",
        anchor: str = "c",
        fill: str = "black",
        font: str = "Helvetica 10",
    ):
        x_y = self._scale_coords((x_coord, y_coord))
        created_text = self.canvas.create_text(
            *x_y, font=font, fill=fill, anchor=anchor, text=text
        )
        return created_text

    def set_coords(self, element_id, x_y):
        x_y = self._scale_coords(x_y)
        canvas_coords = self.canvas.coords(element_id, x_y)
        self.update()
        return canvas_coords

    def delete(self, element_id: int) -> bool:
        self.canvas.delete(element_id)
        self.update()
        return True

    def set_text_value(self, element_id: int, text: str):
        self.canvas.itemconfig(element_id, text=text)
        self.update()

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def update(self):
        """updates the canvas containing gui"""
        self.tk_gui.update()



from casymda.visualization.canvas.scaled_canvas import ScaledCanvas

ANCHOR_DICT = {
    "c": [0.5, 0.5],
    "nw": [0, 0],
    "n": [0.5, 0],
    "ne": [1, 0],
    "e": [1, 0.5],
    "se": [1, 1],
    "s": [0.5, 1],
    "sw": [0, 1],
    "w": [0, 0.5],
}

UPDATED_KEY = "__UPDATED__"


class WebCanvas(ScaledCanvas):
    """
    json-serializable flat dictionary canvas (implements the scaled canvas).
    dictionary is flat so that it can be shared between multiple processes,
    e.g. when writing to the canvas from the simulation, and consuming the information from a browser via a web-server.
    """

    def __init__(self, dictionary, width, height, scale=1.0):
        super().__init__(scale)
        self.dict = dictionary
        self.width = width
        self.height = height

        self.photo_lookup = {}

        self.element_id_counter = 0

        self.dict[UPDATED_KEY] = set()  # set of all updated element_ids

    def load_image_file(self, path):
        potential_key = path
        if potential_key in self.photo_lookup:
            return self.photo_lookup[potential_key]
        self.element_id_counter += 1
        self.dict[self.element_id_counter] = {
            "factor": self.scale,
            "path": path,
            "type": "photo",
        }
        self.update_updated(self.element_id_counter)
        self.photo_lookup[potential_key] = self.element_id_counter
        return self.element_id_counter

    def create_image(self, x_coord, y_coord, image_file, anchor="c"):
        x_coord, y_coord = self._scale_coords((x_coord, y_coord))
        self.element_id_counter += 1
        self.dict[self.element_id_counter] = {
            "x": x_coord,
            "y": y_coord,
            "anchor": ANCHOR_DICT[anchor],
            "type": "image",
            "photo_id": image_file,
            "path": self.dict[image_file]["path"],
            "factor": self.dict[image_file]["factor"],
        }
        self.update_updated(self.element_id_counter)
        return self.element_id_counter

    def create_text(
        self, x_coord, y_coord, text="", anchor="se", fill="black", font="Helvetica 16"
    ):
        x_coord, y_coord = self._scale_coords((x_coord, y_coord))
        self.element_id_counter += 1
        self.dict[self.element_id_counter] = {
            "x": x_coord,
            "y": y_coord,
            "anchor": ANCHOR_DICT[anchor],
            "type": "text",
            "text": text,
            "fill": fill,
            "font_family": font.split(" ")[0],
            "font_size": int(font.split(" ")[1]),
        }
        self.update_updated(self.element_id_counter)
        return self.element_id_counter

    def delete(self, element_id):
        if element_id in self.dict:
            del self.dict[element_id]
            return True
        return False

    def set_coords(self, element_id, x_y):
        x_coord, y_coord = self._scale_coords(x_y)
        if element_id in self.dict:
            entry = self.dict[element_id]
            entry["x"] = x_coord
            entry["y"] = y_coord
            self.dict[element_id] = entry
            self.update_updated(element_id)
            return True
        return False

    def set_text_value(self, element_id, text=""):
        if element_id in self.dict:
            entry = self.dict[element_id]
            entry["text"] = text
            self.dict[element_id] = entry
            self.update_updated(element_id)
            return True
        return False

    def get_width(self):
        return self.width

    def get_height(self):
        return self.height

    def update_updated(self, id: int):
        updated_set = set(self.dict[UPDATED_KEY])
        updated_set.add(id)
        self.dict[UPDATED_KEY] = updated_set  # trigger process synchro



from ctypes import c_bool
from multiprocessing import Value
from time import monotonic as time, sleep

from simpy.core import EmptySchedule, Environment, Infinity


class SyncedFloat:
    """ holds a float value which can be shared between threads and used as a factor for a casymda ChangeableFactorRealtimeEnvironment """

    def __init__(self, synced: Value):
        """can be instanciated via factory class method"""
        self.synced: Value = synced

    def __truediv__(self, other):
        return self.synced.value / other

    def __rtruediv__(self, other):
        return other / self.synced.value

    def __mul__(self, other):
        return other * self.synced.value

    def __rmul__(self, other):
        return self.__mul__(other)

    def set_value(self, value: float) -> float:
        self.synced.value = value
        return self.synced.value

    @classmethod
    def _create_factor_instance(cls, factor=1.0) -> "SyncedFloat":
        value = Value("d", factor)
        return SyncedFloat(value)


class ChangeableFactorRealtimeEnvironment(Environment):
    """Realtime environment without strict-mode, but with possibility to set the factor dynamically"""

    pause_sleep_time = 0.25

    def __init__(
        self,
        initial_time=0,
        factor=SyncedFloat._create_factor_instance(),
        should_run=Value(c_bool, True),
    ):
        """factor can be obtained via 'get_factor_instance' classmethod.
        also works when factor is"""
        self.factor: SyncedFloat = factor
        self.should_run: Value = should_run
        super().__init__(initial_time)

    def step(self):
        """ modified step function without strict-mode but working with changing factors"""

        if not self.should_run.value:
            sleep(self.pause_sleep_time)
            return

        evt_time = self.peek()

        if evt_time is Infinity:
            raise EmptySchedule()

        sleep_time = (evt_time - self.now) * self.factor

        if sleep_time > 0.01:
            sleep(sleep_time)

        super().step()

    @classmethod
    def create_factor_instance(cls, factor=1.0) -> SyncedFloat:
        """ forwards to SyncedFloat factory method """
        return SyncedFloat._create_factor_instance(factor=factor)

    @classmethod
    def create_should_run_instance(cls) -> Value:
        return Value(c_bool, True)



import datetime
from typing import List, Optional
from casymda.blocks.entity import Entity

from casymda.visualization.canvas.scaled_canvas import ScaledCanvas


class EntityAnimation:
    """class to hold animation related info of an entity"""

    def __init__(self, entity, canvas_image, image_file, x_y):
        self.entity = entity
        self.canvas_image = canvas_image
        self.image_file = image_file
        self.x_y = x_y


class EntityVisualizer:
    """ visualizes entities on a canvas when notified """

    def __init__(
        self,
        canvas: ScaledCanvas,
        background_image_path: str = "",
        default_entity_icon_path: str = "",
    ):

        self.canvas = canvas
        self.background_image_path = background_image_path
        self.default_entity_icon_path = default_entity_icon_path

        self.background_image_file = self.canvas.load_image_file(
            self.background_image_path
        )

        self.background_image = self.canvas.create_image(
            0, 0, self.background_image_file, anchor="nw"
        )

        time_text = "Time:  " + str(0)
        # (scale absolute coordinates back)
        self.time_label: int = self.canvas.create_text(
            int((self.canvas.get_width() - 5) / self.canvas.scale),
            int((self.canvas.get_height() - 5) / self.canvas.scale),
            font="Helvetica 10",
            fill="black",
            anchor="se",
            text=time_text,
        )

        self.entity_animations: List[EntityAnimation] = []

    def animate(self, entity: Entity, x: float, y: float, current_time: float):
        """ change entity position """

        # find / create entity animation
        entity_anim = self._find_or_create_entity_animation(entity, x, y)

        entity_icon = entity_anim.canvas_image
        # set position
        self.canvas.set_coords(entity_icon, (x, y))
        entity_anim.x_y = (x, y)
        self.updatetime_label(current_time)

    def updatetime_label(self, now: float):
        """updates the own time_label with the given time"""
        text = "Time:  " + str(datetime.timedelta(seconds=int(now)))
        self.canvas.set_text_value(self.time_label, text=text)

    def create_entity_animation(self, entity: Entity, x_y):
        """creates entity animation"""

        # possibility to look for entity file path if defined
        if entity.process_animation_icon_path is not None:
            file_path = entity.process_animation_icon_path
        else:
            file_path = self.default_entity_icon_path
        image_file = self.canvas.load_image_file(file_path)
        icon = self.canvas.create_image(*x_y, image_file=image_file, anchor="c")

        entity_animation = EntityAnimation(entity, icon, image_file, x_y)
        self.entity_animations.append(entity_animation)

        return entity_animation

    def destroy(self, entity: Entity):
        """destroy entity animation"""
        entity_animation = self._find_entity_animation(entity)
        if entity_animation is not None:
            self.canvas.delete(entity_animation.canvas_image)
            self.entity_animations.remove(entity_animation)
            entity_animation = None

    def update_entity_icon(self, entity: Entity):
        """ recreate entity icon at current position if existing """
        anim = self._find_entity_animation(entity)
        if anim is not None:
            x_y = anim.x_y
            self.destroy(entity)
            self.create_entity_animation(entity, x_y)

    def _find_entity_animation(self, entity: Entity) -> Optional[EntityAnimation]:
        return next((x for x in self.entity_animations if x.entity is entity), None)

    def _find_or_create_entity_animation(
        self, entity: Entity, x: float, y: float
    ) -> "EntityAnimation":
        existing = self._find_entity_animation(entity)
        return (
            self.create_entity_animation(entity, (x, y))
            if existing is None
            else existing
        )



"""setuptools setup"""

from setuptools import find_packages, setup

VERSION = "0.2.26"

try:
    with open("readme.md") as f:
        long_description = f.read()
except FileNotFoundError:
    long_description = ""


setup(
    name="casymda",
    url="https://github.com/fladdimir/casymda",
    author="FFC",
    author_email="fladdi.mir@gmx.de",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={
        "casymda.visualization.state_icons": ["*.png"],
        "casymda.visualization.web_server": ["*.html", "*.js"],
    },
    version=VERSION,
    install_requires=[
        "simpy",
        "xmltodict",
        "Pillow",
        "flask",
        "flask-cors",
        "black==19.3b0",
        "numpy",
        "networkx",
    ],
    zip_safe=False,
    license="MIT",
    description="Simple DES modeling and simulation"
    + " based on SimPy, BPMN, and pixi.js",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
)



"""named resource"""
from simpy import Resource


class NamedResource(Resource):
    """simpy resource with name"""

    def __init__(self, env, name, capacity=1, xy=None, ways=None):
        super().__init__(env, capacity=capacity)

        self.name = name
        self.xy = xy
        self.ways = ways



"""sink"""
from casymda.blocks.block_components import VisualizableBlock


class Sink(VisualizableBlock):
    """sink; yields 0 timeout event before no further processing is started"""

    def __init__(self, env, name, xy=None, ways=None):
        super().__init__(env, name, xy=xy, ways=ways)

    def process_entity(self, entity):
        yield self.env.timeout(0)

        entity.time_of_last_arrival = self.env.now
        self.on_enter(entity)
        self.overall_count_in += 1
        self.entities.append(entity)
        self.block_resource.release(entity.block_resource_request)
        self.on_exit(entity, None)

    def actual_processing(self, entity):
        """not called in this special block"""



import flatbuffers

from . import Anchor, Canvas, CanvasPosition, Element, Image, Photo, Text
from .ElementsUnion import ElementsUnion


def serialize(state: dict):

    num_elements = len(state)
    fb = flatbuffers.Builder(1024)  # default decisive?

    elements = []
    i: str
    e: dict
    for i, e in state.items():

        if e["type"] == "photo":
            path = fb.CreateString(e["path"])
            Photo.PhotoStart(fb)
            Photo.PhotoAddFactor(fb, e["factor"])
            Photo.PhotoAddPath(fb, path)
            elementBody = Photo.PhotoEnd(fb)
            elementType = ElementsUnion.Photo

        elif e["type"] == "image":
            path = fb.CreateString(e["path"])
            Image.ImageStart(fb)
            Image.ImageAddAnchor(fb, Anchor.CreateAnchor(fb, *e["anchor"]))
            Image.ImageAddFactor(fb, e["factor"])
            Image.ImageAddPath(fb, path)
            Image.ImageAddPhotoId(fb, e["photo_id"])
            Image.ImageAddPosition(
                fb, CanvasPosition.CreateCanvasPosition(fb, e["x"], e["y"])
            )
            elementBody = Image.ImageEnd(fb)
            elementType = ElementsUnion.Image

        elif e["type"] == "text":
            fill = fb.CreateString(e["fill"])
            font = fb.CreateString(e["font_family"])
            text = fb.CreateString(e["text"])
            Text.TextStart(fb)
            Text.TextAddAnchor(fb, Anchor.CreateAnchor(fb, *e["anchor"]))
            Text.TextAddFill(fb, fill)
            Text.TextAddFontFamily(fb, font)
            Text.TextAddFontSize(fb, e["font_size"])
            Text.TextAddPosition(
                fb, CanvasPosition.CreateCanvasPosition(fb, e["x"], e["y"])
            )
            Text.TextAddText(fb, text)
            elementBody = Text.TextEnd(fb)
            elementType = ElementsUnion.Text

        Element.ElementStart(fb)
        Element.ElementAddId(fb, int(i))
        Element.ElementAddElementType(fb, elementType)
        Element.ElementAddElement(fb, elementBody)
        element = Element.ElementEnd(fb)
        elements.append(element)

    Canvas.CanvasStartContentVector(fb, num_elements)
    for elem in elements:
        fb.PrependUOffsetTRelative(elem)
    content = fb.EndVector(num_elements)

    Canvas.CanvasStart(fb)
    Canvas.CanvasAddContent(fb, content)
    canvas = Canvas.CanvasEnd(fb)

    fb.Finish(canvas)
    binary = fb.Output()

    return binary



import datetime
import time
from typing import Dict, List, Tuple

import casymda.visualization.state_icons.state_icons_pathfinder as state_icon_pathfinder
from casymda.blocks.block_components import VisualizableBlock
from casymda.blocks.block_components.state import States
from casymda.blocks.block_components.visualizer_interface import BlockVisualizer
from casymda.visualization.canvas.scaled_canvas import ScaledCanvas
from casymda.visualization.entity_visualizer import EntityVisualizer


class ProcessVisualizer(EntityVisualizer, BlockVisualizer):
    """
    visualizes block processes at a given canvas.
    extends EntityVisualizer by BlockVisualizer animation behavior (e.g. for entity flows between blocks)
    """

    def __init__(
        self,
        canvas: ScaledCanvas,
        flow_speed: float = 200,
        background_image_path: str = "",
        default_entity_icon_path: str = "",
    ):

        self.flow_speed = flow_speed  # speed of entity movement in pixels per second

        super().__init__(canvas, background_image_path, default_entity_icon_path)
        self.block_animations: List[BlockAnimation] = []

    def animate_block(self, block: VisualizableBlock, queuing_direction_x=-20):
        # creates new or updates existing animation
        queuing = block.queuing
        block_anim = self._find_or_create_block_anim(block)

        # update text
        if len(block.entities) > 5:  # don"t provide too much details
            content_list_str = "Entities:  " + str(len(block.entities))
        else:
            content_list_str = "\n".join(
                [
                    " - ".join(
                        [
                            entity.name,
                            str(
                                datetime.timedelta(
                                    seconds=int(entity.time_of_last_arrival)
                                )
                            ),
                        ]
                    )
                    for entity in block.entities
                ]
            )
        self.canvas.set_text_value(block_anim.content_list, text=content_list_str)

        self.updatetime_label(block.env.now)

        # update entity animations as well if they should queue
        if queuing:
            x_coord, y_coord = block.xy_position
            queuing_offset_x = 0
            for entity in block.entities:
                self._animate_flow_x_y(
                    x_coord,
                    y_coord,
                    x_coord + queuing_offset_x,
                    y_coord,
                    entity,
                    skip_flow=True,
                )
                queuing_offset_x += queuing_direction_x

    def animate_entity_flow(
        self, entity, from_block: VisualizableBlock, to_block: VisualizableBlock
    ):

        self.updatetime_label(from_block.env.now)

        if to_block is None:
            self.destroy(entity)  # e.g. at the sink
            return

        ways = from_block.ways
        skip_flow = False  # might be moved to block if interesting

        from_x, from_y = from_block.xy_position
        to_x, to_y = to_block.xy_position

        # if waypoints exist, animate flow between waypoints first and update
        # fx, fy
        if ways is not None and to_block.name in ways:
            for way in from_block.ways[to_block.name]:
                self._animate_flow_x_y(
                    from_x, from_y, *way, entity, skip_flow=skip_flow
                )
                from_x, from_y = way

        self._animate_flow_x_y(from_x, from_y, to_x, to_y, entity, skip_flow=skip_flow)

    def _animate_flow_x_y(self, from_x, from_y, to_x, to_y, entity, skip_flow):
        """ change entity position via optional movement animation """

        # find / create entity animation
        entity_anim = self._find_or_create_entity_animation(entity, from_x, from_y)
        entity_icon = entity_anim.canvas_image

        # make sure starting position is as expected
        self.canvas.set_coords(entity_icon, (from_x, from_y))
        self._move_icon(
            from_x, from_y, to_x, to_y, entity_icon, self.flow_speed, skip_flow
        )

        entity_anim.x_y = (to_x, to_y)  # update known position

    def _move_icon(self, from_x, from_y, to_x, to_y, icon, flow_speed, skip_flow):
        """ only animate flow if not skipped """
        if skip_flow:
            self.canvas.set_coords(icon, (to_x, to_y))
            return

        distance = ((to_x - from_x) ** 2 + (to_y - from_y) ** 2) ** 0.5
        if distance == 0:
            return  # target already reached

        # wallclock time needed for distance
        needed_time = distance / float(flow_speed)
        fps = 30  # aninmations per second
        # distance / number of animations = distance per move
        remaining_x_dist = to_x - from_x
        remaining_y_dist = to_y - from_y
        x_per_frame = (remaining_x_dist) / (needed_time * fps)
        y_per_frame = (remaining_y_dist) / (needed_time * fps)

        # animation loop
        while (abs(remaining_x_dist) + abs(remaining_y_dist)) > 0:
            # do not move further than necessary
            x_move = (
                remaining_x_dist
                if abs(remaining_x_dist) < abs(x_per_frame)
                else x_per_frame
            )
            y_move = (
                remaining_y_dist
                if abs(remaining_y_dist) < abs(y_per_frame)
                else y_per_frame
            )

            remaining_x_dist -= x_move
            remaining_y_dist -= y_move

            current_x = to_x - remaining_x_dist
            current_y = to_y - remaining_y_dist

            self.canvas.set_coords(icon, (current_x, current_y))
            time.sleep(1 / float(fps))  # sleep until next frame

    def create_block_animation(self, block: VisualizableBlock):
        """currently only consists of updated list of entities,
        contents list"s position to slightly below the center"""
        x_coord, y_coord = block.xy_position
        x_coord -= 20
        y_coord += 20
        content_list = self.canvas.create_text(
            x_coord,
            y_coord,
            font="Helvetica 10",
            fill="black",
            anchor="nw",
            text="",
        )

        block_animation = BlockAnimation(block, content_list)
        self.block_animations.append(block_animation)
        return block_animation

    def _find_or_create_block_anim(self, block: VisualizableBlock):
        """creates new or updates existing animation"""
        block_animation = next(
            (x for x in self.block_animations if x.block is block), None
        )  # existing
        if block_animation is None:
            block_animation = self.create_block_animation(block)  # new
        return block_animation

    def change_block_state(
        self, block: VisualizableBlock, state: States, new_value: bool
    ):

        block_animation = self._find_or_create_block_anim(block)
        if new_value is True:
            state_icon = StateIcon(
                state.value,
                block.xy_position,
                state_icon_pathfinder.get_file_path(state),
                self.canvas,
            )
            block_animation.state_icons[state.value] = state_icon
        else:
            if state.value in block_animation.state_icons:
                self.canvas.delete(block_animation.state_icons[state.value].icon)
                del block_animation.state_icons[state.value]


class StateIcon:
    """icon to visualize the block's state"""

    OFFSETS: Dict[str, Tuple[int, int]] = {
        States.busy.value: (38, -25),
        States.empty.value: (38, -25),
        States.blocked.value: (38, 0),
    }

    def __init__(
        self,
        state_value: str,
        x_y: Tuple[int, int],
        path_to_icon: str,
        canvas: ScaledCanvas,
    ):
        self.state_value = state_value
        self.x_y: List[int] = list(map(sum, zip(x_y, self.OFFSETS[state_value])))
        self.photo = canvas.load_image_file(path_to_icon)
        self.icon = canvas.create_image(
            self.x_y[0], self.x_y[1], image_file=self.photo, anchor="c"
        )


class BlockAnimation:
    """class to hold block animation"""

    def __init__(self, block, content_list, icon=None, text=None):
        self.block = block
        self.content_list = content_list
        self.icon = icon
        self.text = text
        self.state_icons: Dict[str, StateIcon] = {}



from enum import Enum
from typing import Callable


class States(Enum):
    """ common block states """

    empty = "empty"  # "default", if all other states' count is 0
    busy = "busy"  # counted
    blocked = "blocked"  # by its successor, counted


class StateManager:
    """
    a block is able to be in multiple states at a time.
    counts inc/dec events for each state and evaluates.
    """

    def __init__(self, on_state_change_callback: Callable[[States, bool], None]):
        self._counted_states = [States.busy, States.blocked]  # could be parameter
        self._default_state = States.empty  # could be parameter
        self._state_count = {state.value: 0 for state in self._counted_states}
        self._current_states = {state.value: False for state in States}
        self._on_state_change_callback = on_state_change_callback

    def increment_state_count(self, state):
        self._change_state_count(state, 1)

    def decrement_state_count(self, state):
        self._change_state_count(state, -1)

    def _change_state_count(self, state, change):
        self._state_count[state.value] += change
        self.evaluate_state_count()

    def evaluate_state_count(self):
        for state in self._counted_states:
            if self._state_count[state.value] > 0:
                self._set_new_state(state, True)
            else:
                self._set_new_state(state, False)

        if sum(self._state_count.values()) == 0:
            self._set_new_state(self._default_state, True)
        else:
            self._set_new_state(self._default_state, False)

    def _set_new_state(self, state, new_value):
        if self._current_states[state.value] != new_value:
            self._on_state_change_callback(state, new_value)
            self._current_states[state.value] = new_value
